VARYING vec3 vNormal;
VARYING vec3 vViewVec;

void MAIN()
{
    METALNESS = 1.0;
    ROUGHNESS = 0.25;
    FRESNEL_POWER = 5.0;
    CLEARCOAT_FRESNEL_POWER = 2.0;
    CLEARCOAT_FRESNEL_SCALE = 1.0;
    CLEARCOAT_FRESNEL_BIAS = 0.5;
    CLEARCOAT_AMOUNT = 1.0;
    CLEARCOAT_ROUGHNESS = 0.1;
    CLEARCOAT_NORMAL.y *= 1.5;
    CLEARCOAT_NORMAL = normalize(CLEARCOAT_NORMAL);
}

void AMBIENT_LIGHT()
{
    DIFFUSE += TOTAL_AMBIENT_COLOR;
}
