// Copyright (C) 2019 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
// Qt-Security score:significant reason:default


#include "qssgruntimerenderlogging_p.h"

QT_BEGIN_NAMESPACE

Q_LOGGING_CATEGORY(GL_ERROR, "QSSG.gl_error")
Q_LOGGING_CATEGORY(INVALID_PARAMETER, "QSSG.invalid_parameter")
Q_LOGGING_CATEGORY(INVALID_OPERATION, "QSSG.invalid_operation")
Q_LOGGING_CATEGORY(OUT_OF_MEMORY, "QSSG.out_of_memory")
Q_LOGGING_CATEGORY(INTERNAL_ERROR, "QSSG.internal_error")
Q_LOGGING_CATEGORY(PERF_WARNING, "QSSG.perf_warning")
Q_LOGGING_CATEGORY(PERF_INFO, "QSSG.perf_info")
Q_LOGGING_CATEGORY(WARNING, "QSSG.warning")

QT_END_NAMESPACE
